#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

#include "bblas.h"
#include "batched_blas_test_common.h"

#define typ float

static double getTime(){
  struct timeval t;
  gettimeofday(&t,NULL);
  return t.tv_sec + (double) t.tv_usec*1e-6;
}

int main(int argc, char *argv[])
{
  int i,j,l;
  int nsize,gsize,group_count,*group_size,*group_head;
  bblas_enum_t layout;
  bblas_enum_t *uplo;
  bblas_enum_t *trans;
  int *n,*k,*lda,*ldc;
  typ *alpha,*beta;
  typ **a;
  typ **c,**c_temp;
  double ts,te;
  double error,max_error;
  int try_number;
  int total_batch_count;
  int max_nsize;
  int a_size,c_size;
  int i_try;
  
  if(argc < 4){
    printf("usage : ./(load module) <try_number>  <group_size (random base) > <n (random base)>\n");
    exit(1);
  }

  try_number = atoi(argv[1]);
  group_count = 1;
  gsize = atoi(argv[2]);
  nsize = atoi(argv[3]);

  printf("name= %s, try_number= %d, group_count= %d, group_size= %d, nsize= %d\n", argv[0], try_number, group_count, gsize, nsize);

  srand48(123L);

  group_size = (int *)malloc(sizeof(int) * group_count);
  group_head = (int *)malloc(sizeof(int) * group_count);
  trans = malloc(sizeof(bblas_enum_t) * group_count);
  uplo = malloc(sizeof(bblas_enum_t) * group_count);
  n = (int *)malloc(sizeof(int) * group_count);
  k = (int *)malloc(sizeof(int) * group_count);
  lda = (int *)malloc(sizeof(int) * group_count);
  ldc = (int *)malloc(sizeof(int) * group_count);
  alpha = (typ *)malloc(sizeof(typ) * group_count);
  beta = (typ *)malloc(sizeof(typ) * group_count);

  for(i_try = 0; i_try < try_number; i_try++){
    // set group size
    total_batch_count = 0;
    for(i = 0; i < group_count; i++){
      group_size[i] = random_int((int)((1.0-VAL_RANGE)*gsize),gsize);
      total_batch_count += group_size[i];
    }
    // set group head
    group_head[0] = 0;
    for(i = 1; i < group_count; i++){
      group_head[i] = group_head[i - 1] + group_size[i - 1];
    }

    a = (typ **)malloc(sizeof(typ *) * total_batch_count);
    c = (typ **)malloc(sizeof(typ *) * total_batch_count);
    c_temp = (typ **)malloc(sizeof(typ *) * total_batch_count);
  
    // set parameter and value
    layout = r_layout_();
    for(i = 0; i < group_count; i++){
      trans[i] = r_transpose_c_();
      uplo[i] = r_uplo_();
      n[i] = random_int((int)((1.0-VAL_RANGE)*nsize),nsize);
      k[i] = random_int((int)((1.0-VAL_RANGE)*nsize),nsize);
      max_nsize = MAX_2(n[i],k[i]);
      lda[i] = random_int(max_nsize, (int)(max_nsize*(1.0+VAL_RANGE)));
      ldc[i] = random_int(max_nsize, (int)(max_nsize*(1.0+VAL_RANGE)));
      alpha[i] = drand48();
      beta[i] = drand48();
      for(j = 0; j < group_size[i]; j++){
        if((layout == BblasColMajor && trans[i] == BblasNoTrans) ||
           (layout == BblasRowMajor && (trans[i] == BblasTrans || trans[i] == BblasConjTrans)))
          a_size = lda[i] * k[i];
        else
          a_size = lda[i] * n[i];
        c_size = ldc[i] * n[i];
        a[group_head[i]+j] = (typ *)malloc(sizeof(typ) * 2 * a_size);
        c[group_head[i]+j] = (typ *)malloc(sizeof(typ) * 2 * c_size);
        c_temp[group_head[i]+j] = (typ *)malloc(sizeof(typ) * 2 * c_size);
        for(l = 0; l < 2 * a_size; l++){
          a[group_head[i]+j][l] = drand48();
        }
        for(l = 0; l < 2 * c_size; l++){
          c_temp[group_head[i]+j][l] = c[group_head[i]+j][l] = drand48();
        }
      }
    }
  
    //Set info
    int info_option =  BblasErrorsReportNone;
    if( 5 < argc ){
      int info_no = atoi(argv[4]);
      switch (info_no) {
      case 1 :
	info_option =  BblasErrorsReportAll;
	break;
      case 2 :
	info_option =  BblasErrorsReportGroup;
	break;
      case 3 :
	info_option =  BblasErrorsReportNone;
	break;
      default:
	info_option =  BblasErrorsReportNone;
	break;	
      }
    }

    int info_size;
    switch (info_option) {
    case BblasErrorsReportAll :
      info_size = total_batch_count +1;
      break;
    case BblasErrorsReportGroup :
      info_size = group_count +1;
      break;
    case BblasErrorsReportAny :
      info_size   = 1;
      info_option = BblasErrorsReportNone;
      break;
    case BblasErrorsReportNone :
      info_size = 1;
      break;
    default :
      info_size   = 1;
      info_option = BblasErrorsReportNone;
      break;
    }

    int *info = (int*) malloc((size_t)info_size*sizeof(int))  ;
    info[0] = info_option;
    // batched
    ts = getTime();
    blas_cherk_batchf(group_size[0],layout,*uplo,*trans,*n,*k,*alpha,(const bblas_complex32_t **)a,*lda,*beta,(bblas_complex32_t **)c,*ldc,info);
    te = getTime();
    printf("batched= %e [s],", te - ts);
    free(info);

    // serial
    ts = getTime();
    for(i = 0; i < group_count; i++){
      for(j = 0; j < group_size[i]; j++){
        cblas_cherk(layout_cblas( layout ) , uplo_cblas( uplo[i] ) , transpose_cblas( trans[i] ) , n[i], k[i], alpha[i], a[group_head[i]+j], lda[i],
          beta[i], c_temp[group_head[i]+j], ldc[i]);
      }
    }
    te = getTime();
    printf("serial= %e [s],", te - ts);
    // check
    max_error = 0.0;

    for(i = 0; i < group_count; i++){
      c_size = ldc[i] * n[i];
      for(j = 0; j < group_size[i]; j++){
        for(l = 0; l < c_size; l++){
          error = sqrt(SQR(c_temp[group_head[i]+j][2*l]-c[group_head[i]+j][2*l])
                       +SQR(c_temp[group_head[i]+j][2*l+1]-c[group_head[i]+j][2*l+1]));
          if(error > max_error)
            max_error = error;
        }
      }
    }

    printf("max_error= %e\n", max_error);
  
    for(i = 0; i < group_count; i++){
      for(j = 0; j < group_size[i]; j++){
        free(a[group_head[i]+j]);
        free(c[group_head[i]+j]);
        free(c_temp[group_head[i]+j]);
      }
    }
    free(a);
    free(c);
    free(c_temp);
  }
  
  free(group_size);
  free(group_head);
  free(trans);
  free(uplo);
  free(n);
  free(k);
  free(lda);
  free(ldc);
  free(alpha);
  free(beta);
  
}

