#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

#include "batched_blas.h"
#include "batched_blas_test_common.h"

#define typ double

double getTime(){
  struct timeval t;
  gettimeofday(&t,NULL);
  return t.tv_sec + (double) t.tv_usec*1e-6;
}


int main(int argc, char *argv[])
{
  int i,j,l;
  int nsize,gsize,group_count,*group_size,*group_head;
  CBLAS_LAYOUT layout;
  CBLAS_TRANSPOSE *transa;
  int *m,*n,*lda;
  int *incx,*incy;
  typ *alpha,*beta;
  typ **a,**x;
  typ **y,**y_temp;
  double ts,te;
  double error,max_error;
  int try_number;
  int total_batch_count;
  int max_nsize;
  int a_size,x_size,y_size;
  int i_try;
  
  if(argc < 5){
    printf("usage : ./(load module) <try_number> <group_count> <group_size (random base) > <n (random base)>\n");
    exit(1);
  }

  try_number = atoi(argv[1]);
  group_count = atoi(argv[2]);
  gsize = atoi(argv[3]);
  nsize = atoi(argv[4]);
  
  printf("name= %s, try_number= %d, group_count= %d, group_size= %d, nsize= %d\n", argv[0], try_number, group_count, gsize, nsize);

  srand48(123L);

  group_size = (int *)malloc(sizeof(int) * group_count);
  group_head = (int *)malloc(sizeof(int) * group_count);
  transa = malloc(sizeof(CBLAS_TRANSPOSE) * group_count);
  m = (int *)malloc(sizeof(int) * group_count);
  n = (int *)malloc(sizeof(int) * group_count);
  lda = (int *)malloc(sizeof(int) * group_count);
  incx = (int *)malloc(sizeof(int) * group_count);
  incy = (int *)malloc(sizeof(int) * group_count);
  alpha = (typ *)malloc(sizeof(typ) * group_count);
  beta = (typ *)malloc(sizeof(typ) * group_count);


  for(i_try = 0; i_try < try_number; i_try++){
    // set group size
    total_batch_count = 0;
    for(i = 0; i < group_count; i++){
      group_size[i] = random_int((int)((1.0-VAL_RANGE)*gsize),gsize);
      total_batch_count += group_size[i];
    }
    // set group head
    group_head[0] = 0;
    for(i = 1; i < group_count; i++){
      group_head[i] = group_head[i - 1] + group_size[i - 1];
    }

    a = (typ **)malloc(sizeof(typ *) * total_batch_count);
    x = (typ **)malloc(sizeof(typ *) * total_batch_count);
    y = (typ **)malloc(sizeof(typ *) * total_batch_count);
    y_temp = (typ **)malloc(sizeof(typ *) * total_batch_count);

    // set parameter and value
    layout = r_layout_();
    for(i = 0; i < group_count; i++){
      transa[i] = r_transpose_();
      m[i] = random_int((int)((1.0-VAL_RANGE)*nsize),nsize);
      n[i] = random_int((int)((1.0-VAL_RANGE)*nsize),nsize);
      max_nsize = MAX_2(m[i],n[i]);
      lda[i] = random_int(max_nsize, (int)(max_nsize*(1.0+VAL_RANGE)));
      incx[i] = random_int(1, MAX_INC);
      incy[i] = random_int(1, MAX_INC);
      alpha[i] = drand48();
      beta[i] = drand48();
      for(j = 0; j < group_size[i]; j++){
        if(layout == CblasColMajor)
          a_size = lda[i] * n[i];
        else
          a_size = lda[i] * m[i];
        if(transa[i] == CblasNoTrans)
          x_size = (1+(n[i]-1)*incx[i]);
        else
          x_size = (1+(m[i]-1)*incx[i]);
        if(transa[i] == CblasNoTrans)
          y_size = (1+(m[i]-1)*incy[i]);
        else
          y_size = (1+(n[i]-1)*incy[i]);
        a[group_head[i]+j] = (typ *)malloc(sizeof(typ) * a_size);
        x[group_head[i]+j] = (typ *)malloc(sizeof(typ) * x_size);
        y[group_head[i]+j] = (typ *)malloc(sizeof(typ) * y_size);
        y_temp[group_head[i]+j] = (typ *)malloc(sizeof(typ) * y_size);
        for(l = 0; l < a_size; l++){
          a[group_head[i]+j][l] = drand48();
        }
        for(l = 0; l < x_size; l++){
          x[group_head[i]+j][l] = drand48();
        }
        for(l = 0; l < y_size; l++){
          y_temp[group_head[i]+j][l] = y[group_head[i]+j][l] = drand48();
        }
      }
    }

    // batched
    ts = getTime();
    cblas_dgemv_batch(layout, transa, m, n, alpha, (const typ **)a, lda, (const typ **)x, incx, beta, (typ **)y, incy, group_count, group_size);
    te = getTime();
    printf("batched= %e [s],", te - ts);
  
    // serial
    ts = getTime();
    for(i = 0; i < group_count; i++){
      for(j = 0; j < group_size[i]; j++){
        cblas_dgemv(layout, transa[i], m[i], n[i], alpha[i], a[group_head[i]+j], lda[i], x[group_head[i]+j], incx[i], beta[i], y_temp[group_head[i]+j], incy[i]); 
      }
    }
    te = getTime();
    printf("serial= %e [s],", te - ts);

    // check
    max_error = 0.0;
    for(i = 0; i < group_count; i++){
      for(j = 0; j < group_size[i]; j++){
        if(transa[i] == CblasNoTrans)
          y_size = (1+(m[i]-1)*incy[i]);
        else
          y_size = (1+(n[i]-1)*incy[i]);
        for(l = 0; l < y_size; l++){
          error=fabs(y_temp[group_head[i]+j][l] - y[group_head[i]+j][l]);
        if(max_error < error)
            max_error = error;
        }
      }
    }
    printf("max_error= %e\n", max_error);
    for(i = 0; i < group_count; i++){
      for(j = 0; j < group_size[i]; j++){
        free(a[group_head[i]+j]);
        free(x[group_head[i]+j]);
        free(y[group_head[i]+j]);
        free(y_temp[group_head[i]+j]);
      }
    }
    free(a);
    free(x);
    free(y);
    free(y_temp);
  }
  free(group_size);
  free(group_head);
  free(transa);
  free(m);
  free(n);
  free(lda);
  free(incx);
  free(incy);
  free(alpha);
  free(beta);
  
}


