#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

#include "batched_blas.h"
#include "batched_blas_test_common.h"

#define typ double

double getTime(){
  struct timeval t;
  gettimeofday(&t,NULL);
  return t.tv_sec + (double) t.tv_usec*1e-6;
}


int main(int argc, char *argv[])
{
  int i,j;
  int nsize,gsize,group_count,*group_size,*group_head;
  typ **a,**b,**c,**s;
  typ **a_temp,**b_temp,**c_temp,**s_temp;
  double ts,te;
  double error,max_error;
  int try_number;
  int total_batch_count;
  int i_try;
  
  if(argc < 5){
    printf("usage : ./(load module) <try_number> <group_count> <group_size (random base) > <n (random base)>\n");
    exit(1);
  }

  try_number = atoi(argv[1]);
  group_count = atoi(argv[2]);
  gsize = atoi(argv[3]);
  nsize = atoi(argv[4]);
  
  printf("name= %s, try_number= %d, group_count= %d, group_size= %d, nsize= %d\n", argv[0], try_number, group_count, gsize, nsize);

  srand48(123L);

  group_size = (int *)malloc(sizeof(int) * group_count);
  group_head = (int *)malloc(sizeof(int) * group_count);
  
  for(i_try = 0; i_try < try_number; i_try++){
    // set group size
    total_batch_count = 0;
    for(i = 0; i < group_count; i++){
      group_size[i] = random_int((int)((1.0-VAL_RANGE)*gsize),gsize);
      total_batch_count += group_size[i];
    }
    // set group head
    group_head[0] = 0;
    for(i = 1; i < group_count; i++){
      group_head[i] = group_head[i - 1] + group_size[i - 1];
    }
  
    a = (typ **)malloc(sizeof(typ *) * total_batch_count);
    b = (typ **)malloc(sizeof(typ *) * total_batch_count);
    c = (typ **)malloc(sizeof(typ *) * total_batch_count);
    s = (typ **)malloc(sizeof(typ *) * total_batch_count);
    a_temp = (typ **)malloc(sizeof(typ *) * total_batch_count);
    b_temp = (typ **)malloc(sizeof(typ *) * total_batch_count);
    c_temp = (typ **)malloc(sizeof(typ *) * total_batch_count);
    s_temp = (typ **)malloc(sizeof(typ *) * total_batch_count);
    
    for(i = 0; i < group_count; i++){
      for(j = 0; j < group_size[i]; j++){
        a[group_head[i]+j] = (typ *)malloc(sizeof(typ) * 1);
        b[group_head[i]+j] = (typ *)malloc(sizeof(typ) * 1);
        c[group_head[i]+j] = (typ *)malloc(sizeof(typ) * 1);
        s[group_head[i]+j] = (typ *)malloc(sizeof(typ) * 1);
        a_temp[group_head[i]+j] = (typ *)malloc(sizeof(typ) * 1);
        b_temp[group_head[i]+j] = (typ *)malloc(sizeof(typ) * 1);
        c_temp[group_head[i]+j] = (typ *)malloc(sizeof(typ) * 1);
        s_temp[group_head[i]+j] = (typ *)malloc(sizeof(typ) * 1);
	a[group_head[i]+j][0] = a_temp[group_head[i]+j][0] = drand48();
	b[group_head[i]+j][0] = b_temp[group_head[i]+j][0] = drand48();
	c[group_head[i]+j][0] = c_temp[group_head[i]+j][0] = drand48();
	s[group_head[i]+j][0] = s_temp[group_head[i]+j][0] = drand48();
      }
    }

    // batched
    ts = getTime();
    cblas_drotg_batch(a, b, c, s, group_count, group_size);
    te = getTime();
    printf("batched= %e [s],", te - ts);
  
    // serial
    ts = getTime();
    for(i = 0; i < group_count; i++){
      for(j = 0; j < group_size[i]; j++){
        cblas_drotg(a_temp[group_head[i]+j], b_temp[group_head[i]+j], c_temp[group_head[i]+j], s_temp[group_head[i]+j]);
      }
    }
    te = getTime();
    printf("serial= %e [s],", te - ts);

    max_error = 0.0;
    for(i = 0; i < group_count; i++){
      for(j = 0; j < group_size[i]; j++){
        error = (a[group_head[i]+j][0] - a[group_head[i]+j][0]);
        if(max_error < error)
          max_error = error;
        error = (b[group_head[i]+j][0] - b[group_head[i]+j][0]);
        if(max_error < error)
          max_error = error;
        error = (c[group_head[i]+j][0] - c[group_head[i]+j][0]);
        if(max_error < error)
          max_error = error;
        error = (s[group_head[i]+j][0] - s[group_head[i]+j][0]);
        if(max_error < error)
          max_error = error;
      }
    }
    printf("max_error= %e\n", max_error);
    
    for(i = 0; i < group_count; i++){
      for(j = 0; j < group_size[i]; j++){
        free(a[group_head[i]+j]);
        free(b[group_head[i]+j]);
        free(c[group_head[i]+j]);
        free(s[group_head[i]+j]);
        free(a_temp[group_head[i]+j]);
        free(b_temp[group_head[i]+j]);
        free(c_temp[group_head[i]+j]);
        free(s_temp[group_head[i]+j]);
      }
    }
    free(a);
    free(b);
    free(c);
    free(s);
    free(a_temp);
    free(b_temp);
    free(c_temp);
    free(s_temp);
    
  }
  free(group_size);
  free(group_head);
  
}
